﻿var esprima = require('esprima'),
    escodegen = require('escodegen'),
    traverse = require('ast-traverse'),
    astUtils = require('esprima-ast-utils'),
    md5 = require('md5'),
    path = require('path'),
    getTransientMethodName = require('./transient-method-naming.js').getTransientMethodName,
    guessMethodName = require('./guess-method-name.js').guessMethodName;

function formatLoc(loc) {
    return formatLocToArray(loc).join(',');
}

function formatLocToArray(loc) {
    return [loc.line, loc.column];
}

function calculateFunctionNodeSigHash(node) {
    if (!node.params || node.params.length == 0) return "";
    var paramsStr = node.params.map(escodegen.generate).join(',');
    var hash = md5(paramsStr);
    return hash;
}

function calculateFunctionNodeHash(node) {
    var clonedNode = astUtils.clone(node);
    //calculate body hash
    traverse(clonedNode, {
        pre: function (node, parent, prop, idx) {
            if (node == clonedNode) return;
            if (node.type == "FunctionDeclaration" || node.type == "FunctionExpression") {
                node.body = { type: "BlockStatement", body: [] };
                node.params = [];
                node.id = { name: '_' };
            }
        }
    });

    var generated = escodegen.generate(clonedNode);
    var hash = md5(generated);
    return hash;
}

/**
 * Creates a new Uint8Array based on two different ArrayBuffers
 * source: https://gist.github.com/72lions/4528834
 * @private
 * @param {ArrayBuffers} buffer1 The first buffer.
 * @param {ArrayBuffers} buffer2 The second buffer.
 * @return {ArrayBuffers} The new ArrayBuffer created out of the two.
 */
function removeHashbangPrefix(bytes) {
    function concatArrays(buffer1, buffer2) {
        var tmp = new Uint8Array(buffer1.byteLength + buffer2.byteLength);
        tmp.set(new Uint8Array(buffer1), 0);
        tmp.set(new Uint8Array(buffer2), buffer1.byteLength);
        return tmp;
    };

    var HASHBANG_BYTES = [35, 33];
    var LINE_COMMENT_BYTES = [47, 47];
    if (bytes.length > 2 && bytes[0] == HASHBANG_BYTES[0] && bytes[1] == HASHBANG_BYTES[1]) { //hashbang
        bytes = concatArrays(new Uint8Array(LINE_COMMENT_BYTES), bytes);//insert two slashes to make that line a comment
    }
    return bytes;
}

function getMethodName(node, parent) {
    var mn = node.id && node.id.name;
    if (!mn) {
        mn = guessMethodName(node, parent) || "(Anonymous)";
    }
    return mn;
}

function fixRelPath(p1, p2) {
    try {
        if (!p1 || p1.length == 0) return p1;
        if (!p2 || p2.length == 0) return p1;

        //Make sure both paths are unix-style
        p1 = p1.split('\\').join('/');
        p2 = p2.split('\\').join('/');

        if (p2[0] == '/') return p2; //Absolute path

        var parts1 = p1.split('/');
        parts1.pop(); //remove the filename

        var parts2 = p2.split('/');

        var p3 = parts1.concat(parts2);
        var finalPath = path.normalize(p3.join('/')).replace('\\', '/');
        return finalPath;
    } catch (e) {
        return p2;
    }
}

function generateFileSignature(filename, bytes, sourceMaps) {
    var result = {
        module: { hash: "" },
        methods: []
    };

    filename = filename.replace(/\\/g, '/'); //we want all transients to use unix / instead of windows \

    var ast;

    try {
        bytes = removeHashbangPrefix(bytes);

        //1. Parse the file
        ast = esprima.parse(bytes, {
            loc: true,
            range: true,
            comment: false //No need for those
        });

    } catch (parseException) {
        //TODO: Test
        return { error: "Parse error (1)" };
    }

    try {

        // print AST node types, pre-order (node first, then its children) 
        traverse(ast, {
            pre: function (node, parent, prop, idx) {
                if (node.type == "FunctionDeclaration" || node.type == "FunctionExpression") {
                    var hash = calculateFunctionNodeHash(node);
                    var sigHash = calculateFunctionNodeSigHash(node);
                    var start = formatLocToArray(node.loc.start);
                    var end = formatLocToArray(node.loc.end);
                    var name = getMethodName(node, parent);

                    var originalFilename;
                    var originalName;
                    var mappedPosition;
                    var originalStart;
                    var originalEnd;
                    var originalStartPosition, originalEndPosition;

                    if (sourceMaps) {
                        try {
                            originalStartPosition = sourceMaps.originalPositionFor({ line: start[0], column: start[1] });
                            if (!originalStartPosition || !originalStartPosition.source || originalStartPosition.line === null || originalStartPosition.column === null) {
                                return;
                            }
                            originalFilename = filename;
                            filename = originalStartPosition.source;

                            if (filename.indexOf("webpack:/webpack/") === 0) return; //Embedded resource that has no corresponding source
                            if (filename.indexOf("webpack:///external") === 0) return; //Ignore external module
                            if (filename.indexOf('webpack:///') === 0) {
                                filename = filename.substring(11); //'webpack:///'.length
                            }

                            filename = fixRelPath(originalFilename, filename);

                            if (originalStartPosition.name) {
                                originalName = node.id && node.id.name;
                                name = originalStartPosition.name;
                                if (!node.id)
                                    node.id = { name: originalStartPosition.name };
                                else if (!node.id.name)
                                    node.id.name = originalStartPosition.name;
                            }

                            originalStart = start;
                            start = [originalStartPosition.line, originalStartPosition.column];

                            originalEndPosition = sourceMaps.originalPositionFor({ line: end[0], column: end[1] - 1 });
                            originalEnd = end;
                            end = [originalEndPosition.line, originalEndPosition.column];

                        } catch (e) {
                            console.log('Error calculating source maps:' + e);
                        }
                    }

                    var hasExplicitName = (node.id && node.id.name);
                    var uniqueName = [name, filename, start.join(',')].join('@');

                    result.methods.push({
                        name: name, //Method that can be presented to the user
                        isAnonymous: !hasExplicitName,
                        start: start,
                        end: end,
                        hash: hash,
                        sigHash: sigHash,
                        uniqueName: uniqueName,
                        filename: filename,
                        originalFilename: originalFilename,
                        originalStart: originalStart,
                        originalEnd: originalEnd,
                        originalName: originalName
                    })
                }
            }
        });
    }
    catch (traverseException1) {
        return { error: 'Parse error(2)' }
    }

    try {
        var bodyHash = calculateFunctionNodeHash(ast);
        result.module.hash = bodyHash;
    } catch (traverseException2) {
        return { error: "Parse error (3)" };
    }

    return result;
}

module.exports = generateFileSignature;