﻿var glob = require("glob"),
    Promise = require('promise'),
    path = require('path'),
    fs = require('fs'),
    fileSignature = require('./file-signature.js'),
    methodKeyCreator = require('./method-key.js'),
    uuid = require('uuid'),
    //statuses = require('../statuses.js'),
    util = require('util'),
    debuglog = util.debuglog ? util.debuglog('sl') : require('../alt-debug-log.js').debuglog('sl'),
    parser = require("gitignore-parser"),
    guessMethodName = require('./guess-method-name.js').guessMethodName,
    sourceMap = require("source-map");

function generateNewMappingForFile(filename, fullFilename) {
    var fileSig = {
        filename: filename,
        //status: statuses.FILE_ADDED, // "removed", "modified", "moved", "unchanged"
        methods: []
    };
    filename = filename.replace(/\\/g, '/');
    var oldContent = fs.readFileSync(fullFilename);
    var sourceMaps = readSourceMaps(fullFilename);
    var sig = fileSignature(filename, oldContent, sourceMaps);
    if (!sig || sig.error) {
        //fileSig.status = statuses.FILE_CORRUPTED;
        return null;//fileSig;
    }
    fileSig.hash = sig.module.hash;

    var mapping = {};
    sig.methods.forEach(function (m) {
        var wholeKey = [methodKeyCreator(m.start, m.name, m.isAnonymous), m.hash, new Date().valueOf()].join('|');
        var newId = uuid(wholeKey);
        fileSig.methods.push({
            id: newId,
            isAnonymous: m.isAnonymous,
            hash: m.hash,
            sigHash: m.sigHash,
            position: m.start,
            endPosition: m.end,
            name: m.name,
            uniqueName: m.uniqueName,
            filename: m.filename,
            originalFilename: m.originalFilename,
            originalStart: m.originalStart,
            originalEnd: m.originalEnd,
            originalName: m.originalName
        })
    });
    return fileSig;
}

function readSourceMaps(fullFilename){
	if (!fullFilename) return;
	var sourceMapsFilename = fullFilename+'.map';
	var contents,rawSourceMapJsonData,consumer;
	try{
		contents = fs.readFileSync(sourceMapsFilename).toString();		
	}
	catch(e) {
		//Sourcemap file not found
		return null;		
	}
	
	try{
		rawSourceMapJsonData = JSON.parse(contents);
	} catch(e){
		console.error('Ignoring invalid source map: '+sourceMapsFilename);
		return null;
	}

	try{
		console.log('Procssing source maps from file: '+sourceMapsFilename);
		consumer = new sourceMap.SourceMapConsumer(rawSourceMapJsonData);
		return consumer;
	} catch(e){
		console.error('Error processing source maps for file: '+sourceMapsFilename);
		return null;
	}
}

function getSlIgnore(workspacePath){
    var pattern;
    try {
        var slIgnoreFilePath = path.resolve(workspacePath, '.slignore');
        pattern = fs.readFileSync(slIgnoreFilePath, 'utf8');
        try {
            var compiled = parser.compile(pattern); //may throw if the file is invalid/corrupted
            console.log('Using .slignore file ' + slIgnoreFilePath); 
            return compiled;
        } catch(e) {
            console.error('Not using corrupted .slignore file ' + slIgnoreFilePath+". Using default behavior.");            
            throw e;
        }
    }
    catch (e) {
        // default pattern - all js files under test directory and it's sub directories.
        pattern = "test/**.js";
        return parser.compile(pattern); //should not throw again
    }    
}

function javascriptHandler(handlerParams) {
    var opts = handlerParams.buildArguments;
    var slIgnore = getSlIgnore(opts.workspacepath);
    var finalOutput = handlerParams.finalOutput;

    var p = new Promise(function (resolve, reject) {
        finalOutput.files = finalOutput.files || [];
        var files = [];
        var useSlIgnore = true;
        if (handlerParams.buildArguments.specificFiles)
        {
            files = handlerParams.buildArguments.specificFiles;
            useSlIgnore = false;
        }
        else
        {
            files = glob.sync("**/*.js",
            {
                cwd: opts.workspacepath,
                ignore: ["node_modules/**/*.js"]
            });
        }
        files.forEach(function (filename, fileIdx) {
            var fullFilename = path.resolve(opts.workspacepath, filename);

            // If the file is filtered by the .slignore file:
            if(useSlIgnore && slIgnore.denies(filename)) {
                console.log('Excluded [' + (fileIdx + 1) + '/' + files.length + '] ' + fullFilename);
                return; // Ignored by .slignore, skip it.
            }

            console.log('Processing [' + (fileIdx + 1) + '/' + files.length + '] ' + fullFilename);
            
            var fileSig = generateNewMappingForFile(filename, fullFilename);
            //fileSig.methods.forEach(function(m){ console.log(m.originalFilename);});
            //If source maps are used, the methods in fileSig may originate from excluded files
            
            if (!useSlIgnore) {
                var includedMethods = [];
                fileSig.methods.forEach(function(m){
                    if (m.filename) {
                        if (!slIgnore.denies(m.filename)){
                            includedMethods.push(m);
                            delete m.filename;
                        }
                        else {
                            //console.log('Rejecting method from file: '+m.filename);
                        }
                    } else{
                        includedMethods.push(m);
                    }
                });

                fileSig.methods = includedMethods;
            }
            else {
                fileSig.methods.forEach(function(m){ delete m.filename; });
            }

            if (fileSig) finalOutput.files.push(fileSig);
        })
        resolve(true);
    });
    return p;
}

module.exports = javascriptHandler